from importlib_resources import files
from typing import List
from pandas import read_csv, DataFrame

from eis1600.helper.Singleton import singleton

entities_path = files('eis1600.markdown.data').joinpath('entity_tags.csv')


@singleton
class EntityTags:
    __entity_tags_df = None
    __tag_list = None
    __onom_tag_list = None

    def __init__(self) -> None:
        entity_tags_df = read_csv(entities_path)
        EntityTags.__entity_tags_df = entity_tags_df
        EntityTags.__tag_list = entity_tags_df.loc[entity_tags_df['CATEGORY'].notna(), 'TAG'].to_list()
        EntityTags.__onom_tag_list = entity_tags_df.loc[entity_tags_df['CATEGORY'] == 'ONOMASTIC', 'TAG'].to_list()

    @staticmethod
    def get_entity_tags_df() -> DataFrame:
        return EntityTags.__entity_tags_df

    @staticmethod
    def get_entity_tags() -> List[str]:
        return EntityTags.__tag_list

    @staticmethod
    def get_onom_tags() -> List[str]:
        return EntityTags.__onom_tag_list



from os.path import splitext

from eis1600.markdown.markdown_patterns import BIO_CHR_TO_NEWLINE_PATTERN, \
    HEADER_END_PATTERN, \
    MIU_SPLITTER_PATTERN, MIU_UID_TAG_PATTERN, SIMPLE_HEADING_OR_BIO_PATTERN, MISSING_DIRECTIONALITY_TAG_PATTERN, \
    MIU_TAG_AND_TEXT_PATTERN, \
    NEWLINES_CROWD_PATTERN, \
    NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN, ONLY_PAGE_TAG_PATTERN, PAGE_TAG_IN_BETWEEN_PATTERN, \
    PAGE_TAG_ON_NEWLINE_TMP_PATTERN, PAGE_TAG_PATTERN, \
    PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN, UID_TAG_PATTERN, MIU_UID_TAG_AND_TEXT_SAME_LINE_PATTERN, \
    PARAGRAPH_UID_TAG_PATTERN
from eis1600.texts_to_mius.SubIDs import SubIDs
from eis1600.texts_to_mius.UIDs import UIDs
from eis1600.texts_to_mius.check_formatting_methods import check_formatting_before_update_ids


def pre_clean_text(text: str) -> str:
    text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)
    text = NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN.sub('\n\n', text)
    text = text.replace(' \n', '\n')
    text = text.replace('\n ', '\n')
    text = PAGE_TAG_ON_NEWLINE_TMP_PATTERN.sub(r' \1', text)

    return text


def insert_ids(text: str) -> str:
    # disassemble text into paragraphs
    text = text.split('\n\n')
    text_updated = []

    uids = UIDs()

    text_iter = text.__iter__()
    paragraph = next(text_iter)
    prev_p = ''

    sub_ids = None

    # Insert UIDs and EIS1600 tags into the text
    while paragraph is not None:
        next_p = next(text_iter, None)

        if paragraph:
            # Only do this if paragraph is not empty
            if paragraph.startswith('#'):
                uid = uids.get_uid()
                sub_ids = SubIDs(uid)
                # Move content to an individual line
                paragraph = BIO_CHR_TO_NEWLINE_PATTERN.sub(r'\1\n\2', paragraph)
                paragraph = paragraph.replace('#', f'_ء_#={uid}=')
                # Insert a paragraph tag
                heading_and_text = paragraph.splitlines()
                if len(heading_and_text) == 2:
                    paragraph = heading_and_text[0] + f'\n\n_ء_={sub_ids.get_id()}= ::UNDEFINED:: ~\n_ء_ ' + \
                                heading_and_text[1]
                elif len(heading_and_text) > 2:
                    raise ValueError(
                            f'There is a single new line in this paragraph:\n{paragraph}'
                    )
                text_updated.append(paragraph)
            elif '%~%' in paragraph:
                paragraph = f'_ء_={sub_ids.get_id()}= ::POETRY:: ~\n_ء_ ' + '\n_ء_ '.join(paragraph.splitlines())
                text_updated.append(paragraph)
            elif PAGE_TAG_PATTERN.fullmatch(paragraph):
                page_tag = PAGE_TAG_PATTERN.match(paragraph).group('page_tag')
                if PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN.search('\n\n'.join([prev_p, paragraph, next_p])):
                    if text_updated:
                        if text_updated[-1][-1] == ' ':
                            text_updated[-1] += page_tag + ' ' + next_p
                        else:
                            text_updated[-1] += ' ' + page_tag + ' ' + next_p
                        paragraph = next_p
                        next_p = next(text_iter, None)
                elif text_updated:
                    text_updated[-1] += ' ' + page_tag
            elif paragraph.startswith('::'):
                p_pieces = paragraph.splitlines()
                section_header = p_pieces[0]

                if '%~%' in paragraph:
                    paragraph = '\n_ء_ '.join(p_pieces[1:])
                elif len(p_pieces) > 2:
                    raise ValueError(
                            f'There is a single new line in this paragraph:\n{paragraph}'
                    )
                elif len(p_pieces) == 2:
                    paragraph = p_pieces[1]
                else:
                    raise ValueError(
                            'There is an empty paragraph, check with\n'
                            '::\\n[^هسءگؤقأذپيمجثاڤوضآرتنكزفبعٱشىصلدطغإـئظحةچخ_]'
                    )

                paragraph = f'_ء_={sub_ids.get_id()}= {section_header} ~\n_ء_ ' + paragraph
                text_updated.append(paragraph)
            else:
                paragraph = f'_ء_={sub_ids.get_id()}= ::UNDEFINED:: ~\n_ء_ ' + paragraph
                text_updated.append(paragraph)

        prev_p = paragraph
        paragraph = next_p

    # reassemble text
    text = '\n\n'.join(text_updated)

    return text


def update_ids(text: str) -> str:
    # 1. disassemble text into paragraphs
    text = text.split('\n\n')
    text_updated = []

    used_ids = []

    # 2. Collect existing UIDs
    for idx, paragraph in enumerate(text):
        if MIU_UID_TAG_PATTERN.match(paragraph):
            uid = int(MIU_UID_TAG_PATTERN.match(paragraph).group('UID'))
            if uid not in used_ids:
                used_ids.append(uid)
            else:
                # If in the review process an UID was accidentally inserted twice, just remove the second occurrence
                # - this element will get a new UID in the next steps.
                if MIU_UID_TAG_PATTERN.match(paragraph).group(1):
                    # Python returns None for empty capturing groups which messes up the string
                    text[idx] = MIU_UID_TAG_PATTERN.sub('\1', paragraph)
                else:
                    text[idx] = MIU_UID_TAG_PATTERN.sub('', paragraph)

    uids = UIDs(used_ids)

    text_iter = text.__iter__()
    paragraph = next(text_iter)
    prev_p = None

    # 3. Update MIU_UIDs (sub_ids are adjusted in the next step) and fix some structural problems
    while paragraph is not None:
        next_p = next(text_iter, None)

        if paragraph:
            # Only do this if paragraph is not empty
            if SIMPLE_HEADING_OR_BIO_PATTERN.match(paragraph):
                # Move content to an individual line
                paragraph = BIO_CHR_TO_NEWLINE_PATTERN.sub(r'\1\n\2', paragraph)
                uid = uids.get_uid()
                paragraph = paragraph.replace('#', f'_ء_#={uid}=')
                # Insert a paragraph tag
                heading_and_text = paragraph.splitlines()
                if len(heading_and_text) == 2:
                    paragraph = heading_and_text[0] + f'\n\n::UNDEFINED::\n\n_ء_ {heading_and_text[1]}'
                elif len(heading_and_text) > 2:
                    raise ValueError(
                            f'There is a single new line in this paragraph:\n{paragraph}'
                    )
            elif not UID_TAG_PATTERN.match(paragraph):
                if paragraph.startswith('::'):
                    p_pieces = paragraph.splitlines()
                    section_header = p_pieces[0]
                    if '%~%' in paragraph:
                        paragraph = '\n'.join(p_pieces[1:])
                    elif len(p_pieces) > 2:
                        raise ValueError(
                                f'There is a single new line in this paragraph:\n{paragraph}'
                        )
                    elif len(p_pieces) == 2:
                        paragraph = p_pieces[1]
                    else:
                        raise ValueError(
                                'There is an empty paragraph'
                        )
                else:
                    cat = 'POETRY' if '%~%' in paragraph else 'UNDEFINED'
                    section_header = f'::{cat}::'
                    if cat == 'POETRY':
                        paragraph = '\n'.join(paragraph.splitlines())

                paragraph = f'{section_header}\n' + paragraph
            elif MIU_UID_TAG_AND_TEXT_SAME_LINE_PATTERN.match(paragraph):
                paragraph = MIU_UID_TAG_AND_TEXT_SAME_LINE_PATTERN.sub(r'\1\2\n\3', paragraph)
                # Insert a paragraph tag
                heading_and_text = paragraph.splitlines()
                if '%~%' in '\n'.join(heading_and_text[1:]):
                    paragraph = heading_and_text[0] + f'\n\n::POETRY::\n_ء_ ' + '\n'.join(heading_and_text[1:])
                elif len(heading_and_text) > 2:
                    raise ValueError(
                            f'There is a single new line in this paragraph:\n{paragraph}'
                    )
                else:
                    paragraph = heading_and_text[0] + f'\n\n::UNDEFINED::\n_ء_ ' + heading_and_text[1]
            elif MIU_TAG_AND_TEXT_PATTERN.match(paragraph):
                # TODO
                uid = MIU_TAG_AND_TEXT_PATTERN.match(paragraph).group('UID')
                sub_ids = SubIDs(uid)
                # Insert a paragraph tag
                heading_and_text = paragraph.splitlines()
                if '%~%' in '\n'.join(heading_and_text[1:]):
                    paragraph = heading_and_text[0] + f'\n\n::POETRY::\n_ء_ ' + '\n'.join(heading_and_text[1:])
                elif len(heading_and_text) > 2:
                    raise ValueError(
                            f'There is a single new line in this paragraph:\n{paragraph}'
                    )
                else:
                    paragraph = heading_and_text[0] + f'\n\n::UNDEFINED::\n_ء_ ' + heading_and_text[1]

            if ONLY_PAGE_TAG_PATTERN.fullmatch(paragraph):
                # Add page tags to previous paragraph if there is no other information contained in the current
                # paragraph
                page_tag = ONLY_PAGE_TAG_PATTERN.match(paragraph).group('page_tag')
                if PAGE_TAG_IN_BETWEEN_PATTERN.search('\n\n'.join([prev_p, paragraph, next_p])):
                    if text_updated:
                        next_p_text = next_p.split('::UNDEFINED::\n_ء_ ')[1]
                        if text_updated[-1][-1] == ' ':
                            text_updated[-1] += page_tag + ' ' + next_p_text
                        else:
                            text_updated[-1] += ' ' + page_tag + ' ' + next_p_text
                        paragraph = next_p
                        next_p = next(text_iter, None)
                elif text_updated:
                    text_updated[-1] += ' ' + page_tag
            elif not PARAGRAPH_UID_TAG_PATTERN.fullmatch(paragraph):
                # Do not add empty paragraphs to the updated text
                text_updated.append(paragraph)

            prev_p = paragraph
            paragraph = next_p

    # 4. Update sub_ids
    text = '\n\n'.join(text_updated)
    mius = MIU_SPLITTER_PATTERN.split(text)[1:]     # First element is None because the regex matches the text beginning
    uids = [elem for idx, elem in enumerate(mius) if idx % 2 == 0]
    mius = [elem for idx, elem in enumerate(mius) if idx % 2 == 1]

    text_updated = []

    for uid, miu in zip(uids, mius):
        miu = PARAGRAPH_UID_TAG_PATTERN.sub(r'::\g<cat>::', miu)
        sub_ids = SubIDs(uid)
        paragraphs = miu.split('\n\n')

        for paragraph in paragraphs:
            if paragraph.startswith('::'):
                p_pieces = paragraph.splitlines()
                section_header = p_pieces[0]
                if '%~%' in paragraph:
                    paragraph = '\n'.join(p_pieces[1:])
                elif len(p_pieces) > 2:
                    raise ValueError(
                            f'There is a single new line in this paragraph:\n{paragraph}'
                    )
                elif len(p_pieces) == 2:
                    paragraph = p_pieces[1]
                else:
                    raise ValueError(
                            'There is an empty paragraph, check with\n'
                            '::\\n[^هسءگؤقأذپيمجثاڤوضآرتنكزفبعٱشىصلدطغإـئظحةچخ_]'
                    )
                paragraph = f'_ء_={sub_ids.get_id()}= {section_header} ~\n' + paragraph
            elif not MIU_UID_TAG_PATTERN.match(paragraph):
                paragraph = f'_ء_={sub_ids.get_id()}= ::UNDEFINED:: ~\n' + paragraph

            text_updated.append(paragraph)

    # 5. Reassemble text
    text = '\n\n'.join(text_updated)
    return text


def add_ids(infile: str) -> None:
    """Insert UIDs and EIS1600 tags into EIS1600TMP file and thereby convert it to EIS1600 format.


    :param str infile: Path of the file to convert.
    :return None:
    """
    file_path, file_ext = splitext(infile)

    if file_ext == '.EIS1600':
        # EIS1600 file extension
        with open(infile, 'r+', encoding='utf8') as infile_h:
            text = infile_h.read()

            header_and_text = HEADER_END_PATTERN.split(text)
            header = header_and_text[0] + header_and_text[1]
            text = header_and_text[2].lstrip('\n')  # Ignore new lines after #META#Header#End#
            text = pre_clean_text(text)
            text = MISSING_DIRECTIONALITY_TAG_PATTERN.sub('\g<1>_ء_ \g<2>', text)

            check_formatting_before_update_ids(infile, text)
            text = update_ids(text)

            final = header + '\n\n' + text
            if final[-1] != '\n':
                final += '\n'

            infile_h.seek(0)
            infile_h.write(final)
            infile_h.truncate()
    else:
        # EIS1600TMP file extension
        with open(infile, 'r', encoding='utf8') as infile_h:
            text = infile_h.read()

        header_and_text = HEADER_END_PATTERN.split(text)
        header = header_and_text[0] + header_and_text[1]
        text = header_and_text[2].lstrip('\n')  # Ignore new lines after #META#Header#End#
        text = pre_clean_text(text)

        text = insert_ids(text)

        final = header + '\n\n' + text
        if final[-1] != '\n':
            final += '\n'

        outfile = infile.replace('TMP', '')
        with open(outfile, 'w', encoding='utf8') as outfile_h:
            outfile_h.write(final)



from typing import List

from pandas import notna


def aggregate_STFCON_classes(st_list: list, fco_list: list) -> List[str]:
    label_dict = {
        "B-FAMILY": "F",
        "I-FAMILY": "F",
        "B-CONTACT": "C",
        "I-CONTACT": "C",
        "B-OPINION": "O",
        "I-OPINION": "O",
        "O": "",
        "_": "",
        "B-TEACHER": "T",
        "I-TEACHER": "T",
        "B-STUDENT": "S",
        "I-STUDENT": "S",
        "I-NEUTRAL": "X",
        "B-NEUTRAL": "X",
    }
    aggregated_labels = []
    for a, b in zip(st_list, fco_list):
        labels = f"{label_dict.get(a, '')}{label_dict.get(b, '')}"
        if a == b:
            labels = label_dict.get(a, '')
        else:
            labels = labels.replace("X", "")
        aggregated_labels.append(labels)
    return aggregated_labels


def merge_ner_with_person_classes(ner_labels, aggregated_stfco_labels):
    merged_labels = []
    for a, b in zip(ner_labels, aggregated_stfco_labels):
        if notna(a) and a[:2] == "ÜP":
            postfix = "X"
            if b.strip() != "":
                postfix = b.strip()
            a = a + postfix
        merged_labels.append(a)
    return merged_labels


def merge_ner_with_toponym_classes(ner_labels: List[str], toponym_labels: List[str]) -> List[str]:
    merged_labels = []
    for a, b in zip(ner_labels, toponym_labels):
        if notna(a) and a[:2] == 'ÜT':
            if b:
                merged_labels.append(b)
            else:
                merged_labels.append(a + 'X')
        else:
            merged_labels.append(a)
    return merged_labels



from logging import ERROR, Formatter
from functools import partial
from sys import argv, exit
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600OrEIS1600TMPAction
from p_tqdm import p_uimap
from tqdm import tqdm

from eis1600.helper.chunking import split_file
from eis1600.helper.logging import setup_logger
from eis1600.repositories.repo import TEXT_REPO, get_ready_and_double_checked_files
from eis1600.texts_to_mius.subid_methods import add_ids


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to insert UIDs in EIS1600TMP file(s) and thereby converting them to final EIS1600 
            file(s).
-----
Give a single EIS1600 or EIS1600TMP file as input.

Run without input arg to batch process all double-checked and ready files from the OpenITI_EIS1600_Texts directory.
'''
            )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600 or EIS1600TMP file to process',
            action=CheckFileEndingEIS1600OrEIS1600TMPAction
            )
    arg_parser.add_argument(
            '--split',
            metavar="MB",
            type=int,
            default=5,
            help='split files larger or equal to indicated megabytes. '
                 'Default 4 MB. A value of -1 means no chunking.'
            )
    args = arg_parser.parse_args()

    infile = args.input
    debug = args.debug

    if infile:
        add_ids(infile)
        split_file(infile, args.split, debug)
    else:
        files_ready, files_double_checked = get_ready_and_double_checked_files(only_complete=True)
        files = files_ready + files_double_checked

        if not files:
            print(
                    'There are no more EIS1600 files to process'
            )
            exit()

        formatter = Formatter('%(message)s\n\n\n')
        logger = setup_logger('sub_ids', TEXT_REPO + 'sub_ids.log', ERROR, formatter)
        res = []
        count = 0
        if debug:
            print('\nAdd IDs')
            for i, infile in tqdm(list(enumerate(files))):
                print(i+1, infile)
                try:
                    add_ids(infile)
                except ValueError as e:
                    logger.error(f'{infile}\n{e}\n\n')
                    count += 1
            print(f'{len(files) - count}/{len(files)} processed')

            if args.split == -1:
                print("Remove splitted parts of files")
            else:
                print("Split files")
            for i, infile in tqdm(list(enumerate(files))):
                try:
                    split_file(infile, args.split, debug)
                except ValueError as e:
                    logger.error(f'{infile}\n{e}\n\n')
                    count += 1
        else:
            print('\nAdd IDs')
            res += p_uimap(add_ids, files)
            if args.split == -1:
                print("Remove splitted parts of files")
            else:
                print("Split files")
            res += p_uimap(partial(split_file, max_size=args.split, debug=debug), files)

    print('Done')




from pathlib import Path
from typing import Optional, Tuple, Union
from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from logging import Formatter, INFO
from time import process_time, time
from random import shuffle

from p_tqdm import p_uimap
from tqdm import tqdm
from torch import cuda

from eis1600.bio.md_to_bio import bio_to_md
from eis1600.corpus_analysis.text_methods import get_text_as_list_of_mius
from eis1600.helper.logging import setup_logger
from eis1600.helper.parse_range import parse_range
from eis1600.models.ToponymDescriptionModel import ToponymDescriptionModel
from eis1600.processing.preprocessing import get_yml_and_miu_df
from eis1600.processing.postprocessing import merge_tagslists, reconstruct_miu_text_with_tags
from eis1600.repositories.repo import TEXT_REPO, TOPO_REPO, get_ready_and_double_checked_files


def q_annotate_miu(tup: Tuple[str, str, bool]) -> Union[str, None]:
    uid, miu_as_text, analyse_flag = tup
    yml_handler, df = get_yml_and_miu_df(miu_as_text)

    toponym_labels = ToponymDescriptionModel().predict_sentence(df['TOKENS'].fillna('-').to_list())
    if 'B-TOPD' in toponym_labels:
        df['Q'] = bio_to_md(toponym_labels, umlaut_prefix=False)
        df['TAGS_LISTS'] = df.apply(merge_tagslists, key='Q', axis=1)

        updated_text = reconstruct_miu_text_with_tags(df[['SECTIONS', 'TOKENS', 'TAGS_LISTS']])

        return updated_text
    else:
        return None


def routine_per_text(
        infile: str,
        parallel: Optional[bool] = False,
        force: Optional[bool] = False,
        debug: Optional[bool] = False
    ):
    meta_data_header, mius_list = get_text_as_list_of_mius(infile)
    out_path = infile.replace(TEXT_REPO, TOPO_REPO)

    res = []
    error = ''
    if parallel:
        res += p_uimap(q_annotate_miu, mius_list)
    else:
        for idx, tup in tqdm(list(enumerate(mius_list))):
            try:
                res.append(q_annotate_miu(tup))
            except Exception as e:
                uid, miu_as_text, analyse_flag = tup
                error += f'{uid}\n{e}\n\n\n'

    dir_path = '/'.join(out_path.split('/')[:-1])
    Path(dir_path).mkdir(parents=True, exist_ok=True)

    if not all(r is None for r in res):
        with open(out_path, 'w', encoding='utf-8') as fh:
            fh.write('\n\n'.join([r for r in res if r]))

    if error:
        raise ValueError(error)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to q_annotate whole corpus (e.g. for TOPD annotation).'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument('-P', '--parallel', action='store_true')
    arg_parser.add_argument(
            '--range',
            metavar="ini,end",
            type=parse_range,
            help='process file range [i,j] (both are optional)'
    )
    arg_parser.add_argument(
            "--random", "-r",
            action="store_true",
            help="randomise list of files"
    )
    arg_parser.add_argument(
            "--force", "-f",
            action="store_true",
            help="process file regardless if it exist and overwrite it"
    )

    args = arg_parser.parse_args()
    debug = args.debug
    parallel = args.parallel
    force = args.force

    print(f'GPU available: {cuda.is_available()}')

    st = time()
    stp = process_time()

    # Retrieve all double-checked texts
    files_ready, files_double_checked = get_ready_and_double_checked_files()
    infiles = files_ready + files_double_checked

    if not infiles:
        print('There are no EIS1600 files to process')
        exit()

    formatter = Formatter('%(message)s\n\n\n')
    logger = setup_logger('q_analyse', 'q_analyse.log', INFO, formatter)

    if args.range:
        infiles = infiles[args.range[0]:args.range[1]]

    infiles_indexes = list(range(len(infiles)))
    if args.random:
        shuffle(infiles_indexes)

    for i in tqdm(infiles_indexes):
        infile = infiles[infiles_indexes[i]]
        try:
            print(f'[{i}] {infile}')
            routine_per_text(infile, parallel, force, debug)
        except ValueError as e:
            logger.error(f'{infile}\n{e}')

    et = time()
    etp = process_time()

    print('Done')
    print(f'Processing time: {etp - stp} seconds')
    print(f'Execution time: {et - st} seconds')




import os
import sys
from argparse import ArgumentParser

from eis1600.repositories.repo import get_ready_and_double_checked_files


def main():
    arg_parser = ArgumentParser(
        prog=sys.argv[0],
        description="Simple script to show sizes in MB of all input files"
    )
    arg_parser.add_argument(
        "--sort_asc", "-a",
        action="store_true",
        help="sort files by size in ascending order"
    )
    arg_parser.add_argument(
        "--sort_desc", "-d",
        action="store_true",
        help="sort files by size in descending order"
    )
    arg_parser.add_argument(
        "--complete",
        action="store_true",
        help="get complete files, not part files"
    )
    args = arg_parser.parse_args()

    files_ready, files_double_checked = get_ready_and_double_checked_files(only_complete=args.complete)
    infiles = files_ready + files_double_checked

    if not infiles:
        print('There are no EIS1600 files to process')
        sys.exit()

    infiles = [(f, os.path.getsize(f)) for f in infiles]

    if args.sort_asc:
        infiles.sort(key=lambda x: x[1])
    elif args.sort_desc:
        infiles.sort(key=lambda x: x[1], reverse=True)

    for i, (file, size) in enumerate(infiles, 1):
       mb = size >> 20
       i = f"[{i}]"
       print(f"{i:<5} {size:>8} bytes = {mb:>2} MB - {file}")



from typing import Dict, Optional, Tuple

from eis1600.dates.month_methods import month_annotate_miu_text

from eis1600.bio.md_to_bio import bio_to_md
from eis1600.dates.date_methods import date_annotate_miu_text
from eis1600.nlp.annotation_utils import annotate_miu_text, insert_onom_tag, insert_onomastic_tags
from eis1600.nlp.utils import aggregate_STFCON_classes, merge_ner_with_person_classes, merge_ner_with_toponym_classes
from eis1600.processing.postprocessing import merge_tagslists, reconstruct_miu_text_with_tags
from eis1600.processing.preprocessing import get_yml_and_miu_df
from eis1600.yml.yml_handling import add_annotated_entities_to_yml, add_statistics_to_yml


def analyse_miu(tup: Tuple[str, str, bool], debug: Optional[bool] = False) -> Dict:
    """Analysis the miu with our models.

    This methods applies our models to the text of the MIU and thereby runs different analysis (NER, POS, LEMMAS,
    ROOTS, TOPONYMS, PERSONS, ONOMASTICS, DATES). Results are added to the df representation of the MIU. Returns a
    JSON object which contains yml information, as well as the df with all analysis results.
    :param Tuple tup: Params are given as a Tuple: (uid: str, miu_as_text: str, analyse_flag: bool).
    :param bool debug: Optional flag to print debug statements
    :return Dict: JSON representation of the miu, including yml header and analysis results.
    """
    uid, miu_as_text, analyse_flag = tup

    # 1. open miu file and disassemble the file to its parts
    if debug:
        print('1. open miu file and disassemble the file to its parts')
    yml_handler, df = get_yml_and_miu_df(miu_as_text)

    if analyse_flag:
        # 2. annotate NEs, POS and lemmatize. NE are: person + relation(s), toponym + relation, onomastic information
        if debug:
            print('2. annotate NEs, POS and lemmatize. NE are: person + relation(s), toponym + relation, onomastic information')
        df['NER_LABELS'], df['LEMMAS'], df['POS_TAGS'], df['ROOTS'], ST_labels, FCO_labels, \
            df['TOPONYM_LABELS'] = annotate_miu_text(df, debug)

        # 3. convert cameltools labels format to markdown format
        if debug:
            print('3. convert cameltools labels format to markdown format')
        aggregated_stfco_labels = aggregate_STFCON_classes(ST_labels, FCO_labels)
        ner_tags = bio_to_md(df['NER_LABELS'].to_list())  # camel2md_as_list(df['NER_LABELS'].tolist())
        ner_tags_with_person_classes = merge_ner_with_person_classes(ner_tags, aggregated_stfco_labels)
        toponym_labels_md = bio_to_md(df['TOPONYM_LABELS'].to_list(), sub_class=True)
        df['NER_TAGS'] = merge_ner_with_toponym_classes(ner_tags_with_person_classes, toponym_labels_md)

        # 4. annotate dates
        if debug:
            print('4. annotate dates')
        df['DATE_TAGS'] = date_annotate_miu_text(df[['TOKENS']], uid, yml_handler)
        df['MONTH_TAGS'] = month_annotate_miu_text(df[['TOKENS']], uid)

        # 5. insert BONOM and EONOM tags with the pretrained transformer model
        df['ONOM_TAGS'] = insert_onom_tag(df, debug)

        # 6. annotate onomastic information
        df['ONOMASTIC_TAGS'] = insert_onomastic_tags(df, debug)

        # TODO 6. disambiguation of toponyms (same toponym, different places) --> replace ambiguous toponyms flag
        # TODO 9. get frequencies of unidentified entities (toponyms, nisbas)

        # 11. reconstruct the text, populate yml with annotated entities and save it to the output file
        if debug:
            print('Reconstruct the text, populate yml with annotated entities and save it to the output file')
        columns_of_automated_tags = ['DATE_TAGS', 'MONTH_TAGS', 'ONOM_TAGS', 'ONOMASTIC_TAGS', 'NER_TAGS']
        for col in columns_of_automated_tags:
            if col in df.columns:
                df['TAGS_LISTS'] = df.apply(merge_tagslists, key=col, axis=1)
        df_subset = df[['SECTIONS', 'TOKENS', 'TAGS_LISTS']]

        reconstructed_miu_text_with_tags = reconstruct_miu_text_with_tags(df_subset)
        if debug:
            print("  add_annotated_entities_to_yml")
        add_annotated_entities_to_yml(df_subset, yml_handler, uid, reconstructed_miu_text_with_tags)
        if debug:
            print("  add_statistics_to_yml")
        add_statistics_to_yml(df_subset, yml_handler)
        if debug:
            print("Reconstructed text =", reconstructed_miu_text_with_tags)

    # return as JSON Object
    author, text, edition, miu_uid = uid.split('.')
    yml_init = {'author': author, 'text': text, 'edition': edition, 'UID': miu_uid}
    miu_as_json = {'yml': yml_handler.to_json(yml_init), 'df': df.to_json(force_ascii=False, compression=None)}

    return miu_as_json



from sys import argv
import ujson as json
import pandas as pd
from io import StringIO
from pathlib import Path
from functools import partial
from p_tqdm  import p_uimap
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from eis1600.repositories.repo import get_ready_and_double_checked_files, TEXT_REPO, JSON_REPO, RECONSTRUCTED_INFIX
from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600JsonAction
from eis1600.processing.postprocessing import write_updated_miu_to_file
from eis1600.yml.YAMLHandler import YAMLHandler


def reconstruct_file(fpath: str, force: bool = False):

    if fpath.endswith(".json"):
        out_fpath = fpath.replace(".json", f"{RECONSTRUCTED_INFIX}.EIS1600")
    else:
        fpath = fpath.replace(TEXT_REPO, JSON_REPO)
        out_fpath = fpath.replace(".EIS1600", f"{RECONSTRUCTED_INFIX}.EIS1600")
        fpath = fpath.replace('.EIS1600', '.json')

    # do not process file if it's already generated and it should not be overwritten
    if Path(out_fpath).is_file() and not force:
        return

    with open(fpath, "r", encoding="utf-8") as fp, \
         open(out_fpath, "w", encoding="utf-8") as outfp:
        data = json.load(fp)

        yml = data[0]["yml"]
        yml_handler = YAMLHandler(yml)
        df = pd.concat([pd.read_json(StringIO(miu["df"])) for miu in data], ignore_index=True)
        df["TAGS_LISTS"] = None
        write_updated_miu_to_file(outfp, yml_handler, df, forced_re_annotation=True)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description="Script to reconstruct an EIS1600 text file form the json output."
    )
    arg_parser.add_argument(
            'infile', type=str, nargs='?',
            help='json file to process',
            action=CheckFileEndingEIS1600JsonAction
    )
    arg_parser.add_argument(
            '--force', action='store_true',
            help='create file even though it is already created'
    )
    args = arg_parser.parse_args()

    if args.infile:
        reconstruct_file(args.infile)

    else:
        files_ready, files_double_checked = get_ready_and_double_checked_files()
        files = files_ready + files_double_checked

        list(p_uimap(partial(reconstruct_file, force=args.force), files, num_cpus=0.7))

        print(f"Reconstructed {len(files)} files")
        print(f"For each json file in {JSON_REPO} directory, a reconstructed .EIS1600 file has been created.")



from logging import Formatter, INFO

from eis1600.helper.logging import setup_logger
from eis1600.repositories.repo import TEXT_REPO, get_ready_and_double_checked_files
from eis1600.texts_to_mius.check_formatting_methods import check_formatting
from eis1600.texts_to_mius.subid_methods import add_ids


def main():
    """Script to check file formatting based on text selection."""
    ready_files, double_checked_files = get_ready_and_double_checked_files(only_complete=True)

    formatter = Formatter('%(message)s\n\n\n')
    logger = setup_logger('mal_formatted_texts', TEXT_REPO + 'mal_formatted_texts.log', INFO, formatter)

    logger.info('insert_uids')
    print('Insert UIDs into ready texts')

    x = 0
    for i, file in enumerate(ready_files[x:]):
        print(i + x, file)
        try:
            add_ids(file)
        except ValueError as e:
            logger.error(f'{file}\n{e}')

    logger.info('\n\n\ndisassemble_text')
    print('Check formatting for double-checked and ready texts')

    texts = double_checked_files + [r.replace('TMP', '') for r in ready_files]

    count = 0
    x = 0
    for i, text in enumerate(texts[x:]):
        print(i + x, text)
        try:
            check_formatting(text)
        except ValueError as e:
            count += 1
            logger.error(e)
        except FileNotFoundError:
            print(f'Missing: {text}')

    print(f'\n{count}/{len(texts)} texts need fixing\n')

    print('Done')




import os.path

import jsonpickle
from sys import argv
import ujson as json
import pandas as pd
from tqdm import tqdm
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from eis1600.repositories.repo import get_ready_and_double_checked_files, TEXT_REPO, JSON_REPO, COLUMNS, SEP, SEP2
from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600JsonAction


ALL_LABELS = ("SECTIONS", "TOKENS", "TAGS_LISTS", "NER_LABELS", "LEMMAS", "POS_TAGS", "ROOTS", "TOPONYM_LABELS",
              "NER_TAGS", "DATE_TAGS", "MONTH_TAGS", "ONOM_TAGS", "ONOMASTIC_TAGS")


def dump_file(fpath: str, label_list: tuple[str] = ALL_LABELS):

    if not fpath.endswith(".json"):
        fpath = fpath.replace(TEXT_REPO, JSON_REPO)
        fpath = fpath.replace('.EIS1600', '.json')

    structural_data, content_data = [], []
    with open(fpath, "r", encoding="utf-8") as fp:
        data = json.load(fp)

    for miu in data:
        header = miu["yml"]
        uid = header["UID"]

        # get structural data
        for entity in header.keys():
            if entity == "UID":
                continue
            value = header[entity]
            if type(value) in (str, int, bool):
                structural_data.append((uid, entity, value))
            elif type(value) == list:
                for sub_value in value:
                    if type(sub_value) == dict:
                        parsed_sub_value = SEP.join(f"{k}{SEP2}{v}" for k, v in sub_value.items())
                        structural_data.append((uid, entity, parsed_sub_value))
                    elif type(sub_value) == list:
                        parsed_sub_value = SEP.join(sub_value)
                        structural_data.append((uid, entity, parsed_sub_value))
                    else:
                        structural_data.append((uid, entity, sub_value))
            elif type(value) == dict:
                for sub_entity, sub_value in value.items():
                    if type(sub_value) == list:
                        for val in sub_value:
                            structural_data.append((uid, entity, f"{sub_entity}{SEP}{val}"))
                    elif type(sub_value) == dict:
                        # e.g. onomastics elements
                        for sub_sub_ent, sub_sub_val in sub_value.items():
                            structural_data.append((uid, entity, f"{sub_entity}{SEP}{sub_sub_ent}{SEP2}"))
                    else:
                        structural_data.append((uid, entity, f"{sub_entity}{SEP}{sub_value}"))
            else:
                raise ValueError(f'Fatal error dumping data of "{fpath}".')

        # get content data
        miu_df = jsonpickle.decode(miu["df"])
        for entity in label_list:
            if entity not in miu_df:
                continue
            for j, (_, value) in enumerate(miu_df[entity].items(), 1):
                if type(value) == list:
                    value = SEP2.join(value)
                if value:
                    content_data.append((uid, entity, f"{j}{SEP}{value}"))

    fbase, _ = os.path.splitext(fpath)

    content_df = pd.DataFrame(content_data, columns=COLUMNS)
    content_df.to_csv(f"{fbase}_df.tsv", sep="\t", index=False)

    struct_df = pd.DataFrame(structural_data, columns=COLUMNS)
    struct_df.to_csv(f"{fbase}_yml.tsv", sep="\t", index=False)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description="Script to dump all eis1600 corpus into a tsv with the structure and "
                        "another tsv file with the enriched textual information."
    )
    arg_parser.add_argument(
            'infile', type=str, nargs='?',
            help='json file to process',
            action=CheckFileEndingEIS1600JsonAction
    )
    arg_parser.add_argument(
        "--label_list",
        nargs="*",
        default=ALL_LABELS,
        help="entities from content data to add to output. The default is all entities: "
             "SECTIONS, TOKENS, TAGS_LISTS, NER_LABELS, LEMMAS, POS_TAGS, ROOTS, TOPONYM_LABELS, "
             "NER_TAGS, DATE_TAGS, MONTH_TAGS, ONOM_TAGS, ONOMASTIC_TAGS"
    )
    args = arg_parser.parse_args()

    if args.infile:
        dump_file(args.infile, args.label_list)

    else:
        files_ready, files_double_checked = get_ready_and_double_checked_files()
        files = files_ready + files_double_checked

        for fpath in tqdm(files):

            dump_file(fpath, args.label_list)

        print(f"Processed {len(files)} files")
        print(f"For each json file in {JSON_REPO} directory, "
              f"a tsv with the yml data and a tsv with the df data have been generated.")




from typing import List

from openiti.helper.ara import normalize_ara_heavy

ALIF_VARIATIONS = ['ا', 'أ', 'ٱ', 'آ', 'إ']

def normalize_dict(o_set: dict) -> dict:
    n_set = {}

    for key, val in o_set.items():
        n_set[normalize_ara_heavy(key)] = val

    return n_set


def denormalize_list(elem: str) -> List[str]:
    # TODO evaluate if this is enough
    n_list = []
    tmp = [elem]
    if elem.startswith(('أ', 'ٱ', 'آ', 'إ')):
        # alifs
        tmp.extend([alif_variation + elem[1:] for alif_variation in ALIF_VARIATIONS])
    if elem.startswith(('الأ', 'الٱ', 'الآ', 'الإ')):
        #  al + alifs
        tmp.extend(['ال' + alif_variation + elem[3:] for alif_variation in ALIF_VARIATIONS])
    if elem.endswith(('يء', 'ىء', 'ؤ', 'ئ')):
        # hamzas
        tmp.extend([var[:-1] + 'ء' for var in tmp])
    if elem.endswith('ى'):
        # alif maqsura
        tmp.extend([var[:-1] + 'ي' for var in tmp])
    if elem.endswith('ي'):
        # alif maqsura
        tmp.extend([var[:-1] + 'ى' for var in tmp])
    if elem.endswith('ة'):
        # ta marbuta
        tmp.extend([var[:-1] + 'ه' for var in tmp])

    n_list.extend(tmp)

    return n_list



from datetime import date
from functools import partial
from pathlib import Path
from typing import Dict, Optional, Tuple, Union
from sys import argv
from argparse import ArgumentParser, FileType, RawDescriptionHelpFormatter
from re import compile
from json import dump

from pandas import Series

from p_tqdm import p_uimap
from tqdm import tqdm

from eis1600.helper.CheckFileEndingActions import CheckIsDirAction
from eis1600.bio.md_to_bio import md_to_bio
from eis1600.processing.preprocessing import get_yml_and_miu_df
from eis1600.repositories.repo import RESEARCH_DATA_REPO

LABEL_DICT = {'B-TOPD': 0, 'I-TOPD': 1, 'O': 2}
CATS = ['N', 'T']
Q_PATTERN = compile(r'Q(?P<num_tokens>\d+)(?P<cat>[' + ''.join(CATS) + ']*)')
stat = {'NOT REVIEWED': 0, 'REVIEWED': 0, 'REVIEWED2': 0, 'EXCLUDED': 0}


def reconstruct_automated_tag(row) -> str:
    return 'Q' + row['num_tokens']


def get_q_true(file: str, bio_main_class: str, keep_automatic_tags: Optional[bool] = False) \
        -> Tuple[str, Union[Dict, None]]:
    """

    :param file: miu to process.
    :param bio_main_class: tag is composed like this: B-<bio_main_class>, defaults to B-Q.
    :param keep_automatic_tags: Keep Ü-tags, defaults to false.
    :return: Tuple of reviewed status and bio-tags dict.
    """
    with open(file, 'r', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object, keep_automatic_tags)

    stat[yml_handler.reviewed] = stat[yml_handler.reviewed] + 1
    if yml_handler.reviewed != 'REVIEWED':
        return yml_handler.reviewed, None

    s_notna = df['TAGS_LISTS'].loc[df['TAGS_LISTS'].notna()].apply(lambda tag_list: ','.join(tag_list))
    df_true = s_notna.str.extract(Q_PATTERN).dropna(how='all')
    tops = df_true.apply(reconstruct_automated_tag, axis=1)
    tops.name = 'TRUE'

    if not tops.empty:
        df = df.join(tops)
    else:
        return yml_handler.reviewed, None

    bio_tags = md_to_bio(
            df[['TOKENS', 'TRUE']],
            'TRUE',
            Q_PATTERN,
            bio_main_class,
            LABEL_DICT
    )

    return yml_handler.reviewed, bio_tags


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to extract Q-annotations from MIUs and create BIO-training-data.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument(
            'input', type=Path, nargs=1,
            help='Directory which holds the files to process or individual file to annotate',
            action=CheckIsDirAction
    )
    arg_parser.add_argument(
            'out_file', type=FileType('w'), nargs=1,
            help='''Name for the JSON file containing the training-data (without file ending).
            E.G. Q/q_training_data'''
    )
    arg_parser.add_argument(
            'bio_main_class', type=str, nargs='?', default='Q', action='store',
            help='BIO main class, B-<bio_main_class>, defaults to B-Q',
    )

    args = arg_parser.parse_args()
    debug = args.debug
    bio_main_class = args.bio_main_class
    input_df = args.input
    out_file = args.out_file
    keep = True

    mius = input_df.glob('*.EIS1600')

    res = []
    if debug:
        for idx, miu in tqdm(list(enumerate(mius))):
            try:
                res.append(get_q_true(miu, bio_main_class, keep))
            except Exception as e:
                print(idx, miu)
                print(e)
    else:
        res += p_uimap(partial(get_q_true, bio_main_class=bio_main_class, keep_automatic_tags=keep), mius)

    reviewed, bio_dicts = zip(*res)
    bio_dicts = [r for r in bio_dicts if r is not None]

    with open(
            RESEARCH_DATA_REPO + out_file + '_' + date.today().isoformat() + '.json',
            'w',
            encoding='utf-8'
    ) as fh:
        dump(bio_dicts, fh, indent=4, ensure_ascii=False)

    print(Series(reviewed).value_counts())
    print('Done')



from sys import argv
from tqdm import tqdm
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from eis1600.corpus_analysis.analyse_all_on_cluster import routine_per_text
from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600TextAction
from eis1600.repositories.repo import get_all_part_files
from eis1600.helper.part_file_names import get_part_number
from eis1600.json_to_tsv.corpus_dump import dump_file


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to re-annotated files from the online-editor.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument('-P', '--parallel', action='store_true')
    arg_parser.add_argument(
        '--parts', action='store_true',
        help="process all related part files if any"
    )
    arg_parser.add_argument(
            'input', type=str,
            help='EIS1600 text file to annotate.',
            action=CheckFileEndingEIS1600TextAction
    )
    arg_parser.add_argument(
        '--no_tsv',
        action='store_true',
        help='do not make tsv conversion'
    )
    args = arg_parser.parse_args()

    if args.parts:
        for infile in tqdm(sorted(get_all_part_files(args.input), key=get_part_number)):
            if args.debug:
                print(f"** {infile}")
            routine_per_text(infile, parallel=args.parallel, force=True, debug=args.debug)
            if not args.no_tsv:
                dump_file(infile)
    else:
        routine_per_text(args.input, parallel=args.parallel, force=True, debug=args.debug)
        if not args.no_tsv:
            dump_file(args.input)



from typing import Dict, List, Optional, Pattern, Tuple, Union

from eis1600.markdown.EntityTags import EntityTags
from numpy import nan
from pandas import DataFrame, Series


def get_bio_dict(bio_main_class: str, categories: List[str]) -> Dict:
    """

    :param str bio_main_class: String of two letters which are used to indicate the main entity, e.g. 'YY' for dates,
    'TO' for toponyms, etc.
    :param List[str] categories: List of letters used for the categories.
    :return:
    """
    bio = ["B", "I"]

    labels = [bi + "-" + bio_main_class + c for c in categories for bi in bio] + ["O"]
    label_dict = {}

    for i, label in enumerate(labels):
        label_dict[label] = i

    return label_dict


def md_to_bio(
        df: DataFrame, column_name: str, pattern: Pattern, bio_main_class: str, bio_dict: Dict,
        less_categories: Optional[bool] = False, replacements: Optional[Dict] = None
) -> Dict:
    """Parses EIS100 mARkdown tags to BIO labels.

    :param DataFrame df: DataFrame must have these two columns: 'TOKENS' and a second column named <column_name> which
    contains the EIS1600 tags to parse to the BIO labels.
    :param str column_name: Name of the DataFrames column which contains the EIS1600 tags to parse to BIO labels.
    :param Pattern pattern: Pattern must match named capturing groups for: 'num_tokens' and 'cat'.
    :param str bio_main_class: String of two letters which are used to indicate the main entity, e.g. 'YY' for dates,
    'TO' for toponyms, etc.
    :param Dict bio_dict: dictionary whose keys are the BIO labels and the values are integers (see method
    get_label_dict in this file).
    :param less_categories: Flag to indicate combining certain classes into broader categories.
    :param replacements: If less_categories is true, this is a dictionary with the replacement for the categories.
    :return: Dictionary with three entries: 'tokens', a list of the tokens which have been classified; 'ner_tags',
    a list of the numerical representation of the BIO labels assigned to the tokens; 'ner_classes', a list of the
    str representation of the BIO labels assigned to the tokens.
    """
    if any(df[column_name].notna()):
        s_notna = df[column_name].loc[df[column_name].notna()]
        df_matches = s_notna.str.extract(pattern).dropna(how='all')

        if df_matches.empty:
            df['BIO'] = 'O'
        else:
            # This line is needed due to this new bug in pandas where nan is cast to str in str columns, therefore set
            # column explicitly to dtype object so we can call isna() (otherwise == 'nan', which does not make sense)
            df['BIO'] = Series(index=df.index, dtype=object)
            for index, row in df_matches.iterrows():
                processed_tokens = 0
                num_tokens = int(row['num_tokens'])
                cat = row['cat']
                if less_categories and cat in replacements.keys():
                    cat = replacements.get(cat)
                while processed_tokens < num_tokens:
                    if processed_tokens == 0:
                        df.loc[index, 'BIO'] = 'B-' + bio_main_class + cat
                    else:
                        df.loc[index + processed_tokens, 'BIO'] = 'I-' + bio_main_class + cat

                    processed_tokens += 1

            df['BIO'].loc[df['BIO'].isna()] = 'O'
    else:
        df['BIO'] = 'O'

    df['BIO_IDS'] = df['BIO'].apply(lambda bio_tag: bio_dict[bio_tag])
    idcs = df['TOKENS'].loc[df['TOKENS'].notna()].index

    return {
            'tokens': df['TOKENS'].loc[idcs].to_list(),
            'ner_tags': df['BIO_IDS'].loc[idcs].to_list(),
            'ner_classes': df['BIO'].loc[idcs].to_list()
    }


def get_temp_class_es(_label: Union[str, None], sub_class: bool) -> Tuple[str, Union[None, str]]:
    temp_sub_class = None
    if _label[2:] == 'LOC':
        _label = _label.replace('LOC', 'TOX')
        temp_class = _label[2]
    elif _label[2:] in EntityTags().get_onom_tags():
        temp_class = _label[2:]
    elif _label[2:] == 'TOPD':
        temp_class = 'Q'
    else:
        temp_class = _label[2]
    if sub_class:
        temp_sub_class = _label[-1]

    return temp_class, temp_sub_class


def bio_to_md(bio_labels: List[str], sub_class: Optional[bool] = False, umlaut_prefix: Optional[bool] = True) -> List[
    str]:
    """Converts BIO labels to EIS1600 tags.

    Converter method for BIO labels to EIS100 tags. BI labels must follow this pattern: [BI]-[AMPTY].* with
    * [A]ge
    * [M]ISC
    * [P]erson
    * [Q]
    * [T]oponym
    * [Y]ear
    Usually, EIS1600 BIO labels have a three letter code: [YY][BDKP] with YY for year and [BDKP] for the sub-class.
    :param List[str] bio_labels: List containing the BIO label for each token.
    :param bool sub_class: if set to True, last char of BIO label indicates sub-class, e.g. YYB for date with
    sub-class date of birth, defaults to False.
    :param bool umlaut_prefix: if set to False, the md-tags will not be prefixed with 'Ü', defaults to True.
    :return List[str]: List containing the respective EIS1600 tags
    """
    converted_tokens, temp_tokens, temp_class, temp_sub_class = [], [], None, None

    for _label in bio_labels:
        if _label is None:
            converted_tokens.append(nan)
        else:
            # Check if the first letter of the label is 'O' or 'B' because this will terminate the previous entity
            # and 'B' will additionally start a new entity
            if _label[0] in ['O', 'B']:
                if len(temp_tokens):
                    # Generate EIS1600 tag for entity
                    if sub_class:
                        if umlaut_prefix:
                            converted_tokens.append(f"Ü{temp_class}{len(temp_tokens)}{temp_sub_class}")  # e.g. ÜP3X
                        else:
                            converted_tokens.append(f"{temp_class}{len(temp_tokens)}{temp_sub_class}")  # e.g. ÜP3X
                    else:
                        if umlaut_prefix:
                            converted_tokens.append(f"Ü{temp_class}{len(temp_tokens)}")  # e.g. ÜP3
                        else:
                            converted_tokens.append(f"{temp_class}{len(temp_tokens)}")  # e.g. ÜP3
                    # Mask I-tags with nan
                    converted_tokens.extend([nan] * (len(temp_tokens) - 1))
                    temp_tokens = []
                    temp_class = None
                    temp_sub_class = None
                if _label[0] == 'B':
                    # Start new entity
                    temp_class, temp_sub_class = get_temp_class_es(_label, sub_class)
                    temp_tokens = [_label]
                elif _label == 'O':
                    converted_tokens.append(nan)
            elif _label[0] == 'I':
                temp_tokens.append(_label)
                if temp_class is None:
                    temp_class, temp_sub_class = get_temp_class_es(_label, sub_class)
            else:
                converted_tokens.append(nan)

    if len(temp_tokens):
        # Add last entity
        if sub_class:
            converted_tokens.append(f"Ü{temp_class}{len(temp_tokens)}{temp_sub_class}")  # e.g. ÜP3X
        else:
            converted_tokens.append(f"Ü{temp_class}{len(temp_tokens)}")  # e.g. ÜP3
        converted_tokens.extend([nan] * (len(temp_tokens) - 1))

    return converted_tokens




import os.path


def get_part_number(file_name: str) -> int:
    """ get part number from file
    e.g. '0748Dhahabi.TarikhIslam.MGR20180917-ara1_part0016.EIS1600' -> 16
    """
    file_base, _ = os.path.splitext(file_name)
    part_number = file_base.rpartition("_")[-1]
    return int(part_number[4:])



from typing import List, Optional

from numpy import nan

from eis1600.bio.md_to_bio import bio_to_md
from eis1600.models.Lemmatizer import Lemmatizer
from eis1600.models.POSTagger import POSTagger
from eis1600.models.NERModel import NERModel
from eis1600.models.Disambiguator import Disambiguator
from eis1600.models.NasabDetectionModel import NasabDetectionModel
from eis1600.models.OnomsticElementsModel import OnomasticElementsModel
from eis1600.models.FamilyContactOpinionModel import FCOModel
from eis1600.models.StudentTeacherModel import STModel
from eis1600.models.ToponymModel import ToponymModel


def insert_onom_tag(df, debug: Optional[bool] = False) -> List[str]:
    tokens = df['TOKENS'].fillna('-').to_list()
    nasab_tagger = NasabDetectionModel()
    shortend_list_of_tokens = tokens[1:]
    __shortend_list_limit = 120
    if len(tokens) > __shortend_list_limit:
        shortend_list_of_tokens = tokens[1:__shortend_list_limit]
    for idx, t in enumerate(shortend_list_of_tokens):
        if t.strip() == "":
            shortend_list_of_tokens[idx] = "-"
    nasab_labels = nasab_tagger.predict_sentence(shortend_list_of_tokens, debug)
    nasab = ['_']
    nasab_started = False
    for token, label in zip(shortend_list_of_tokens, nasab_labels):
        if label == "B-NASAB":
            # Start a new NASAB
            nasab.append("BONOM")
            nasab_started = True
        elif label == "I-NASAB":
            nasab.append(nan)
        else:
            if nasab_started:
                nasab.append("EONOM")
                nasab_started = False
            else:
                nasab.append(nan)
    if nasab_started:
        nasab[-1] = "EONOM"
    # merge the shortend list
    if len(tokens) > __shortend_list_limit:
        nasab.extend([nan] * (len(tokens) - __shortend_list_limit))
    return nasab


def insert_onomastic_tags(df, debug: Optional[bool] = False):
    onomastic_tagger = OnomasticElementsModel()
    onomastic_tags = [nan] * len(df['TOKENS'])
    start_nasab_id, end_nasab_id = -1, -1

    # Find BNASAB & ENASAB
    for idx, tag in enumerate(df['ONOM_TAGS'].to_list()):
        if "BONOM" == tag:
            start_nasab_id = idx
        elif "EONOM" == tag:
            end_nasab_id = idx
            break

    if 0 < start_nasab_id < end_nasab_id:
        nasab_tokens = df['TOKENS'].fillna('-').to_list()[start_nasab_id:end_nasab_id]
        onomastic_labels = onomastic_tagger.predict_sentence(nasab_tokens, debug)
        ono_tags = bio_to_md(onomastic_labels)

        for i, tag in enumerate(ono_tags):
            onomastic_tags[start_nasab_id + i] = tag

    return onomastic_tags


def annotate_miu_text(df, debug: Optional[bool] = False):
    lemmas, pos_tags, root_tags, ner_tags, st_tags, fco_tags, toponym_tags = ['_'], ['_'], ['_'], ['_'], ['_'], ['_'], ['_']
    section_id, temp_tokens = None, []
    for _section, _token in list(zip(df['SECTIONS'].to_list(), df['TOKENS'].fillna('-').to_list()))[1:]:
        if _section:
            # Start a new section
            if len(temp_tokens) > 0:
                # 1. process the previous section
                lemmas.extend(Lemmatizer().get_lemmas(temp_tokens, debug))
                pos_tags.extend(POSTagger().get_pos(temp_tokens, debug))
                root_tags.extend(Disambiguator().get_roots(temp_tokens, debug))
                ner_tags.extend(NERModel().predict_sentence(temp_tokens, debug))
                st_tags.extend(STModel().predict_sentence(temp_tokens, debug))
                fco_tags.extend(FCOModel().predict_sentence(temp_tokens, debug))
                toponym_tags.extend(ToponymModel().predict_sentence(temp_tokens, debug))

                # 2. reset variables
                section_id, temp_tokens = None, []

        token = _token if _token else '_'
        temp_tokens.append(token)

    if len(temp_tokens) > 0:
        lemmas.extend(Lemmatizer().get_lemmas(temp_tokens, debug))
        pos_tags.extend(POSTagger().get_pos(temp_tokens, debug))
        root_tags.extend(Disambiguator().get_roots(temp_tokens, debug))
        ner_tags.extend(NERModel().predict_sentence(temp_tokens, debug))
        st_tags.extend(STModel().predict_sentence(temp_tokens, debug))
        fco_tags.extend(FCOModel().predict_sentence(temp_tokens, debug))
        toponym_tags.extend(ToponymModel().predict_sentence(temp_tokens, debug))

    return ner_tags, lemmas, pos_tags, root_tags, st_tags, fco_tags, toponym_tags



import sys
import glob
import os.path
from typing import Optional
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from functools import partial
from pathlib import Path
from logging import INFO
from time import process_time, time
from random import shuffle

import jsonpickle
from tqdm import tqdm
from p_tqdm import p_uimap

from torch import cuda

from eis1600.corpus_analysis.miu_methods import analyse_miu
from eis1600.corpus_analysis.text_methods import get_text_as_list_of_mius
from eis1600.json_to_tsv.corpus_dump import dump_file
from eis1600.helper.logging import setup_persistent_logger
from eis1600.helper.parse_range import parse_range
from eis1600.repositories.repo import JSON_REPO, TEXT_REPO, PART_NAME_INFIX, PART_NUM_REGEX, \
                                       get_ready_and_double_checked_files


def routine_per_text(
        infile: str,
        parallel: Optional[bool] = False,
        force: Optional[bool] = False,
        clean_out_dir: Optional[bool] = False,
        debug: Optional[bool] = False,
    ):
    """Entry into analysis routine per text.

    Each text is disassembled into the list of MIUs. Analysis is applied to each MIU. Writes a JSON file containing
    the list of MIUs with their analysis results.
    :param ste infile: EIS1600 text which is analysed.
    :param bool parallel: Parallel flag for parallel processing, otherwise serial processing.
    :param bool force: Do processing even though file already exists.
    :param bool clean_out_dir: When processing all files, clean json output in case there are previous splitting.
        This param will remove all json files in subfolder when file is original or part 1.
    :param bool debug: Debug flag for more console messages.
    """
    out_path = infile.replace(TEXT_REPO, JSON_REPO)
    out_path = out_path.replace('.EIS1600', '.json')

    # do not process file if it's already generated and it should not be overwritten
    if Path(out_path).is_file() and not force:
        return

    meta_data_header, mius_list = get_text_as_list_of_mius(infile)

    res = []
    error = ''
    if parallel:
        res += p_uimap(partial(analyse_miu, debug=debug), mius_list)
    else:
        for idx, tup in tqdm(list(enumerate(mius_list))):
            try:
                res.append(analyse_miu(tup, debug))
            except ValueError as e:
                uid, miu_as_text, analyse_flag = tup
                error += f'{uid}\n{e}\n\n\n'
            except Exception:
                raise

    dir_path, _ = os.path.split(out_path)
    Path(dir_path).mkdir(parents=True, exist_ok=True)

    #FIXME remove original too !!!!!
    # if file is original or part 1, remove previous json files to avoid problem with previous chunkings
    if clean_out_dir and (PART_NAME_INFIX not in out_path or int(PART_NUM_REGEX.search(out_path).group(1)) == 1):
        if os.path.exists(dir_path):
            for json_file in glob.iglob(f"{dir_path}*.json"):
                os.remove(json_file)

    with open(out_path, 'w', encoding='utf-8') as fh:
        jsonpickle.set_encoder_options('json', indent=4, ensure_ascii=False)
        json_str = jsonpickle.encode(res, unpicklable=False)
        fh.write(json_str)

    if error:
        raise ValueError(error)


def main():
    arg_parser = ArgumentParser(
            prog=sys.argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to parse whole corpus to annotated MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument('-P', '--parallel', action='store_true')
    arg_parser.add_argument(
        '--no_tsv',
        action='store_true',
        help='do not make tsv conversion'
    )
    arg_parser.add_argument(
            '--range',
            metavar="ini,end",
            type=parse_range,
            help='process file range [i,j] (both are optional)'
    )
    arg_parser.add_argument(
            "--random", "-r",
            action="store_true",
            help="randomise list of files"
    )
    arg_parser.add_argument(
            "--force", "-f",
            action="store_true",
            help="process file regardless if it exist and overwrite it"
    )

    args = arg_parser.parse_args()
    debug = args.debug
    parallel = args.parallel
    force = args.force

    print(f'GPU available: {cuda.is_available()}')

    st = time()
    stp = process_time()

    # Retrieve all double-checked texts
    files_ready, files_double_checked = get_ready_and_double_checked_files()
    infiles = files_ready + files_double_checked

    if not infiles:
        print('There are no EIS1600 files to process')
        exit()

    logger = setup_persistent_logger('analyse_all_on_cluster', 'analyse_all_on_cluster.log', INFO)

    if args.range:
        infiles = infiles[args.range[0]:args.range[1]]

    infiles_indexes = list(range(len(infiles)))

    if args.random:
        shuffle(infiles_indexes)

    for i in tqdm(infiles_indexes):
        infile = infiles[i]
        print(f"[{i+1}] {infile}")

        try:
            routine_per_text(infile, parallel=parallel, force=force, clean_out_dir=True, debug=debug)
            if not args.no_tsv:
                dump_file(infile)
        except ValueError as e:
            logger.error(f'{infile}\n{e}')
        except Exception as e:
            print(e)
            logger.exception(f'{infile}\n{e}')

    et = time()
    etp = process_time()

    print('Done')
    print(f'Processing time: {etp - stp} seconds')
    print(f'Execution time: {et - st} seconds')



from numpy import nan
from openiti.helper.ara import normalize_ara_heavy
from pandas import DataFrame, Series

from eis1600.dates.Month import Month
from eis1600.dates.date_patterns import MONTHS_NOR, MONTH_IN_CONTEXT_PATTERN
from eis1600.processing.preprocessing import get_tokens_and_tags


def tag_months_fulltext(text: str) -> str:
    text_updated = text

    m = MONTH_IN_CONTEXT_PATTERN.search(text_updated)
    while m:
        if m.group('month'):
            month_str = m.group('month')
            if month_str.startswith('ال'):
                month_str = month_str[2:]
            month_str = normalize_ara_heavy(month_str)
            month = MONTHS_NOR.get(month_str)
            if month == -1:
                # This is not a valid month but still needed in the dict so that the regex for dates matches
                m = MONTH_IN_CONTEXT_PATTERN.search(text_updated, m.end('month'))
            else:
                date = Month(month, len(m.group('month').split()))
                pos = m.start('month')
                text_updated = text_updated[:pos] + date.get_tag() + text_updated[pos:]

                # Recognize next date phrase
                m = MONTH_IN_CONTEXT_PATTERN.search(text_updated, m.end('month') + len(date.get_tag()))
        else:
            # Recognize next date phrase
            m = MONTH_IN_CONTEXT_PATTERN.search(text_updated, m.end('month'))

    return text_updated


def month_annotate_miu_text(ner_df: DataFrame, file: str) -> Series:
    """Annotate dates in the headings and in the MIU text, returns a list of tag per token.

    :param DataFrame ner_df: df containing the 'TOKENS' column.
    :param str file: file name.
    :return Series: List of date tags per token, which can be added as additional column to the df.
    """
    df = ner_df.mask(ner_df == '', None)
    tokens = df['TOKENS'].dropna()
    ar_text = ' '.join(tokens)

    try:
        tagged_text = tag_months_fulltext(ar_text)
    except ValueError as e:
        print(e)
        print(f'Check {file}')

        return Series([nan] * len(df))
    else:
        ar_tokens, tags = get_tokens_and_tags(tagged_text)
        df.loc[df['TOKENS'].notna(), 'MONTH_TAGS'] = tags

        return df['MONTH_TAGS']



from typing import Dict, List, Optional, Tuple
from threading import Lock
from importlib_resources import files

from pandas import read_csv
from transformers import pipeline

from eis1600.helper.Singleton import singleton

poetic_meters_path = files('eis1600.models.data').joinpath('poetic_meters.csv')


@singleton
class PoetryDetectionModel:
    def __init__(self) -> None:
        self.model = pipeline('text-classification', model='CAMeL-Lab/bert-base-arabic-camelbert-mix-poetry')
        self.lock = Lock()
        poetic_meters__df = read_csv(poetic_meters_path)
        self.not_poetry = poetic_meters__df.loc[~poetic_meters__df['POETRY'], 'METER']

    def predict_is_poetry(self, tokens: List[str], debug: Optional[bool] = False) -> Tuple[bool, Dict]:
        with self.lock:
            if debug:
                print('CAMeL-Lab/bert-base-arabic-camelbert-mix-poetry')
            res = self.model(' '.join(tokens))
            res = res[0]
            res['text'] = ' '.join(tokens)
            if res['label'] in self.not_poetry or res['score'] < 0.95:
                return False, res
            else:
                return True, res




import logging

FORMATTER = logging.Formatter('%(message)s')
PERSISTENT_FORMATTER = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')


def setup_logger(
        name: str,
        log_file: str,
        level=logging.INFO,
        formatter=FORMATTER,
        add_stderr: bool = False,
    ) -> logging.Logger:

    handler = logging.FileHandler(log_file, 'w', encoding='utf-8')
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)

    logger.setLevel(level)
    logger.addHandler(handler)
    if add_stderr:
        logger.addHandler(logging.StreamHandler())

    return logger


def setup_persistent_logger(
        name: str,
        log_file: str,
        level=logging.INFO,
        formatter=PERSISTENT_FORMATTER
    ) -> logging.Logger:

    handler = logging.FileHandler(log_file, encoding='utf-8')
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger



from eis1600.models.Model import Model
from eis1600.helper.Singleton import singleton


@singleton
class ToponymModel(Model):

    def __init__(self) -> None:
        super().__init__('EIS1600_Pretrained_Models/camelbert-ca-finetuned_toponyms/')



from typing import Dict, Union

from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from re import compile
from datetime import date
from json import dump

from p_tqdm import p_uimap

from eis1600.markdown.markdown_patterns import PAGE_TAG_PATTERN, SPACES_CROWD_PATTERN
from eis1600.processing.preprocessing import get_tokens_and_tags
from eis1600.repositories.repo import TOPO_TRAINING_REPO

DESCRIPTION_PATTERN = compile('BTOPD((?:\n|.)+?)ETOPD')

BIO_DICT = {'B-TOPD': 0, 'I-TOPD': 1}


def to_bio(entry: str) -> Dict:
    tokens, tags = get_tokens_and_tags(entry)
    bio, bio_ids = [], []
    inside = False
    
    for token, tag in zip(tokens, tags):
        if tag == 'BTOPD':
            inside = True
            bio.append('B-TOPD')
            bio_ids.append(0)
        elif tag == 'ETOPD':
            inside = False
            bio.append('O')
            bio_ids.append(2)
        elif inside:
            bio.append('I-TOPD')
            bio_ids.append(1)
        else:
            bio.append('O')
            bio_ids.append(2)

    return {
            "tokens": tokens,
            "ner_tags": bio_ids,
            "ner_classes": bio
    }


def check_if_entry_is_annotated(entry: str) -> Union[Dict, None]:
    if entry.startswith('$DIC_TOP$') and DESCRIPTION_PATTERN.search(entry):
        return to_bio(SPACES_CROWD_PATTERN.sub(' ', PAGE_TAG_PATTERN.sub('', entry[9:].replace('\n', ' '))))
    else:
        return None


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotate onomastic information in gold-standard MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')

    args = arg_parser.parse_args()
    debug = args.debug

    with open('geographical_dictionaries/data/0900IbnCabdMuncimHimyari.RawdMictar.Shamela0001043-ara1.EIS1600TMP') as fh:
        text = fh.read()

    text = text.split('#META#Header#End#')[1]
    entries = text.split('\n\n# ')
    
    with open('geographical_dictionaries/data/0626YaqutHamawi.MucjamBuldan.Shamela0023735-ara1.EIS1600TMP') as fh:
        text = fh.read()

    text = text.split('#META#Header#End#')[1]
    entries.extend(text.split('\n\n# '))

    res = []
    if debug:
        for entry in entries:
            print(entry)
            res.append(check_if_entry_is_annotated(entry))
    else:
        res += p_uimap(check_if_entry_is_annotated, entries)

    with open(
            TOPO_TRAINING_REPO + 'toponyms_description_training_data_' + date.today().isoformat() + '.json',
            'w',
            encoding='utf-8'
    ) as fh:
        dump([r for r in res if r is not None], fh, indent=4, ensure_ascii=False)

    print('Done')






import os.path
import sys
import ujson as json
from tqdm import tqdm
from itertools import zip_longest
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from eis1600.repositories.repo import get_ready_and_double_checked_files, TEXT_REPO, JSON_REPO
from eis1600.corpus_analysis.text_methods import get_text_as_list_of_mius
from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600TextAction


def compare_mius(infile, debug=False):
    """ return None is no errors, or string with error messages if errors """

    try:
        _, mius_list = get_text_as_list_of_mius(infile)

        original_mius = []
        for miu_block in mius_list:
            _, miu_id = miu_block[0].rsplit(".", 1)
            original_mius.append(miu_id)

        json_file = infile.replace(TEXT_REPO, JSON_REPO)
        json_file = json_file.replace('.EIS1600', '.json')

    except ValueError:
        return f"file {infile} contains errors"

    if not os.path.exists(json_file):
        return f"file {json_file} does not exist"

    with open(json_file, "r", encoding="utf-8") as fp:
        data = json.load(fp)

        json_mius = [miu["yml"]["UID"] for miu in data]

        errors = []
        errors_found = False

        for ori_m, json_m in zip_longest(original_mius, json_mius):

            if ori_m != json_m:

                msg = (f"list of mius do not match!\n"
                       f"    original file = {infile}\n"
                       f"    json file     = {json_file}")

                if debug:
                    if not errors_found:
                        errors.append(msg)
                        errors_found = True
                    errors.append(f"    >> original miu {ori_m} != json miu {json_m}")
                else:
                    return msg

        if errors:
            return "\n".join(errors)


def main():
    arg_parser = ArgumentParser(
            prog=sys.argv[0], formatter_class=RawDescriptionHelpFormatter,
            description="check miu list is complete in json files"
    )
    arg_parser.add_argument(
            'infile', type=str, nargs='?',
            help='EIS1600 or EIS1600TMP file to process',
            action=CheckFileEndingEIS1600TextAction
    )
    arg_parser.add_argument(
        '--ignore_errors',
        action='store_true',
        help='ignore errors when retrieving list of MIUs'
    )
    arg_parser.add_argument(
        '--ignore_missing',
        action='store_true',
        help='do not show warnings for missing json files'
    )
    arg_parser.add_argument(
        '-D',
        '--debug',
        action='store_true'
    )
    args = arg_parser.parse_args()

    if args.infile:
        if errors := compare_mius(args.infile, args.debug):
            print("Comparison was not successful!")
            print(errors)

    else:
        files_ready, files_double_checked = get_ready_and_double_checked_files()
        files = files_ready + files_double_checked

        if not files:
            print('There are no more EIS1600 files to process')
            sys.exit()

        errors = []
        err_count = 0

        for i, infile in tqdm(list(enumerate(files))):
            if args.debug:
                print(f"[{i+1}] {infile}")
            if err := compare_mius(infile, args.debug):
                errors.append(err)
                err_count += 1

        if errors:
            print("Comparison was not successful!")
            print("\n".join(errors))
            print(f"\n{err_count} files failed. "
                  "You can run `check_mius EIS1600_FILE --debug` to see the mismatch of a specific file")
        else:
            print("Everything matches")




from os import remove
from os.path import splitext, exists
from itertools import zip_longest
from eis1600.markdown.markdown_patterns import MIU_UID_TAG_AND_TEXT_SAME_LINE_PATTERN, \
    NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN, NEW_LINE_INSIDE_PARAGRAPH_NOT_POETRY_PATTERN, PARAGRAPH_TAG_MISSING, \
    EMPTY_PARAGRAPH_CHECK_PATTERN, SIMPLE_MARKDOWN, MISSING_DIRECTIONALITY_TAG_PATTERN, SPAN_ELEMENTS, \
    TEXT_START_PATTERN, TILDA_HICKUPS_PATTERN, HEADER_END_PATTERN, CATEGORY_PATTERN, MIU_TAG_PATTERN
from eis1600.markdown.markdown_patterns import SIMPLE_MARKDOWN_TEXT_START_PATTERN
from eis1600.repositories.repo import get_part_filepath


def check_file_for_mal_formatting(infile: str, content: str):
    errors = []

    for line in content.splitlines():
        if m := MIU_TAG_PATTERN.match(line):
            try:
                CATEGORY_PATTERN.search(m.group('category')).group(0)
            except AttributeError:
                errors.append(f"** Category not recognised in \"{m.group(0)}\"")

    # Poetry is still to messed up, do not bother with it for now
    # or POETRY_ATTACHED_AFTER_PAGE_TAG.search(content):

    if not TEXT_START_PATTERN.match(content):
        errors.append('* Text does not start with an MIU tag, check if the preface is tagged as PARATEXT.')
    if m := PARAGRAPH_TAG_MISSING.search(content):
        errors.append(f'* There are missing paragraph tags. Failed at "{m.group(0)}"')
    if m := SIMPLE_MARKDOWN.search(content):
        errors.append(f'* There is simple mARkdown left. Failed at "{m.group(0)}"')
    if m := NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN.search(content):
        errors.append('* There are elements missing the double newline (somewhere the emtpy line is missing). '
                      f'Failed at "{m.group(0)}"')
    if m := MIU_UID_TAG_AND_TEXT_SAME_LINE_PATTERN.search(content):
        errors.append('* There is text on the same line as the start of a biography, fix it by running '
                      f'`ids_insert_or_update` on {infile}. Failed at "{m.group(0)}"')
    if m := TILDA_HICKUPS_PATTERN.search(content):
        errors.append(f'* There is this pattern with tildes: `~\\n~`. Failed at "{m.group(0)}"')
    if m := NEW_LINE_INSIDE_PARAGRAPH_NOT_POETRY_PATTERN.search(content):
        errors.append('* There is a single newline inside a paragraph (somewhere the emtpy line is missing). '
                      f'Failed at "{m.group(0)}"')
    if m := EMPTY_PARAGRAPH_CHECK_PATTERN.search(content):
        errors.append(f'* There are empty paragraphs in the text. Failed at "{m.group(0)}"')
    if m := SPAN_ELEMENTS.search(content):
        errors.append(f'* There are span elements in the text. Failed at "{m.group(0)}"')
    if m := MISSING_DIRECTIONALITY_TAG_PATTERN.search(content):
        errors.append('* There are missing direction tags at the beginning of paragraphs, fix it by running '
                      f'`ids_insert_or_update` on {infile}. Failed at "{m.group(0)}"')
    # if POETRY_ATTACHED_AFTER_PAGE_TAG.search(content):
    #     errors.append('* There is poetry attached to a PageTag (there should be a linebreak instead).')

    if errors:
        errors_str = "\n".join(errors)
        raise ValueError(
            f'Correct the following errors\n'
            f'open -a kate {infile}\n'
            f'kate {infile}\n'
            f'{errors_str}\n\n'
            f'And now run\n'
            f'ids_insert_or_update {infile}\n'
            f'Check if everything is working with\n'
            f'check_formatting {infile}\n'
        )


def check_formatting(infile: str):
    with open(infile, 'r', encoding='utf8') as text:
        header_text = text.read().split('#META#Header#End#\n\n')
        try:
            check_file_for_mal_formatting(infile, header_text[1])
        except ValueError:
            raise


def check_formatting_before_update_ids(infile: str, content: str):
    if not TEXT_START_PATTERN.match(content) and not SIMPLE_MARKDOWN_TEXT_START_PATTERN.match(content) \
            or NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN.search(content) \
            or TILDA_HICKUPS_PATTERN.search(content) \
            or NEW_LINE_INSIDE_PARAGRAPH_NOT_POETRY_PATTERN.search(content) \
            or EMPTY_PARAGRAPH_CHECK_PATTERN.search(content) \
            or SPAN_ELEMENTS.search(content):

        error = ''
        if not TEXT_START_PATTERN.match(content) and not SIMPLE_MARKDOWN_TEXT_START_PATTERN.match(content):
            error += f'\n * Text does not start with Header or PARATEXT, check if the preface is tagged.'
        if m := NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN.search(content):
            error += f'\n * There are elements missing the double newline (somewhere the emtpy line is missing). ' \
                     f'Failed at "{m.group(0)}"'
        if m := TILDA_HICKUPS_PATTERN.search(content):
            error += f'\n * There is this pattern with tildes: `~\\n~`. Failed at "{m.group(0)}"'
        if m := NEW_LINE_INSIDE_PARAGRAPH_NOT_POETRY_PATTERN.search(content):
            error += f'\n * There is a single newline inside a paragraph (somewhere the emtpy line is missing). ' \
                     f'Failed at "{m.group(0)}"'
        if m := EMPTY_PARAGRAPH_CHECK_PATTERN.search(content):
            error += f'\n * There are empty paragraphs in the text. Failed at "{m.group(0)}"'
        if m := SPAN_ELEMENTS.search(content):
            error += f'\n * There are span elements in the text. Failed at "{m.group(0)}"'

        raise ValueError(
            f'Correct the following errors\n'
            f'open -a kate {infile}\n'
            f'kate {infile}\n'
            f'{error}\n\n'
            f'And now run\n'
            f'ids_insert_or_update {infile}\n'
            f'Check if everything is working with\n'
            f'check_formatting {infile}\n'
        )


def check_file_split(infile: str, debug: bool = False):
    with open(infile) as infp:
        original_content = infp.read()
    file_base, file_ext = splitext(infile)

    i = 1
    chunks = []
    header = ""
    part_paths = []  # in case of error they will be removed

    while True:

        if not exists(part_fpath := get_part_filepath(file_base, i, file_ext)):
            break

        part_paths.append(part_fpath)

        with open(part_fpath) as infp:
            part_content = infp.read()

        header_and_text = HEADER_END_PATTERN.split(part_content)
        header = header_and_text[0] + header_and_text[1]
        text = header_and_text[2].lstrip('\n')

        chunks.append(text)
        i += 1

    if original_content == (complete := header + '\n\n' + "\n".join(chunks)):
        # if debug:
        #    print(f"File {infile} has been splitted correctly or has only one part")
        return

    if debug:
        max_lines_print = 50
        print(f"\nError splitting file {infile}. These lines differ (original vs reconstructed):\n")
        for ori, part in zip_longest(original_content.splitlines(True), complete.splitlines(True)):
            if ori != part:
                print(f"{repr(ori):>100}  ≠  {repr(part):>100}")
                max_lines_print -= 1
            if not max_lines_print:
                print("...")
                break

    for fpath in part_paths:
        remove(fpath)

    raise f"splitting of file {infile} failed!"



from typing import Union


class SubIDs:

    def __init__(self, uid: Union[int, str]):
        self.id = 0
        if isinstance(uid, int):
            self.uid_as_str = str(uid)
        else:
            self.uid_as_str = uid

    def get_id(self):
        id_as_str = str(self.id)
        while len(id_as_str) < 8:
            id_as_str = '0' + id_as_str

        self.id += 10

        return self.uid_as_str + '-' + id_as_str



import sys
import logging
from tqdm import tqdm
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600OrEIS1600TMPAction
from eis1600.helper.logging import setup_logger
from eis1600.repositories.repo import TEXT_REPO, get_ready_and_double_checked_files
from eis1600.texts_to_mius.check_formatting_methods import check_formatting


def main():
    arg_parser = ArgumentParser(
            prog=sys.argv[0], formatter_class=RawDescriptionHelpFormatter,
            description=''
    )
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600 or EIS1600TMP file to process',
            action=CheckFileEndingEIS1600OrEIS1600TMPAction
    )
    arg_parser.add_argument(
            '--parts', action='store_true',
            help='check part files instead of original in case file is splitted',
    )

    args = arg_parser.parse_args()

    infile = args.input

    if infile:
        check_formatting(infile)
    else:
        files_ready, files_double_checked = get_ready_and_double_checked_files(only_complete=not args.parts)
        files = files_ready + files_double_checked

        formatter = logging.Formatter('%(message)s\n\n\n')
        logger = setup_logger(
            name='mal_formatted_texts',
            log_file=TEXT_REPO + 'mal_formatted_texts.log',
            level=logging.INFO,
            formatter=formatter,
            add_stderr=True
        )
        print('Check formatting for double-checked and ready texts')

        count = 0
        for text in tqdm(files):
            try:
                check_formatting(text)
            except ValueError as e:
                count += 1
                logger.error(e)
            except FileNotFoundError:
                print(f'Missing: {text}')

        logger.info(f'\n\n\n{count}/{len(files)} files need fixing')

    print('Done')



from threading import Lock
from typing import List, Optional

from camel_tools.disambig.mle import MLEDisambiguator
from eis1600.helper.Singleton import singleton


@singleton
class Disambiguator:
    def __init__(self) -> None:
        self.model = MLEDisambiguator.pretrained()
        self.lock = Lock()

    def get_roots(self, tokens: List[str], debug: Optional[bool] = False) -> List[str]:
        with self.lock:
            if debug:
                print('Roots')
            return [d.analyses[0].analysis['root'] for d in self.model.disambiguate(tokens)]



from re import search
from os import remove
from typing import List
from os.path import splitext, getsize, exists
from itertools import groupby

from eis1600.repositories.repo import PART_NAME_INFIX, get_part_filepath
from eis1600.texts_to_mius.check_formatting_methods import check_file_split
from eis1600.markdown.markdown_patterns import HEADER_END_PATTERN, MISSING_DIRECTIONALITY_TAG_PATTERN, \
    FIRST_LEVEL_HEADING_PATTERN
from eis1600.texts_to_mius.subid_methods import pre_clean_text


def resize_chunks(chunks: List[str], n: int = 400_000) -> List[str]:
    """ If there are chunks with size smaller than n, merge them in groups to avoid having too many tiny files.

    :param list chunks: chunks of text.
    :param int n: soft maximum number of characters per chunk.
    :return list: modified chunking.
    """
    new_chunks = []
    aux = ""
    for chunk in chunks:
        if len(aux) > n:
            new_chunks.append(aux)
            aux = ""
        if aux:
            aux += "\n\n"
        aux += chunk
    if aux:
        # if the last piece of text is too small, add it to the last
        if len(aux) < n // 2 and new_chunks:
            new_chunks[-1] += f"\n\n{aux}"
        else:
            new_chunks.append(aux)
    return new_chunks


def split_file(infile: str, max_size: int, debug: bool = False):
    """ Break files larger or equal MB indicated in max_size into parts so that they can be processed separately.
    All chunks will share the same header.


    :param str infile: Path of the file to split.
    :param int max_size: Maximum size of file in megabytes. If file is larger, it is splitted in parts.
        I max_size is -1, do not split.
    :param bool debug: show warnings.
    :return None:
    """
    if PART_NAME_INFIX in infile:
        return

    file_path, file_ext = splitext(infile)

    # remove previous splitting
    i = 1
    while exists(old_part_file := get_part_filepath(file_path, i, file_ext)):
        remove(old_part_file)
        i += 1

    if max_size == -1:
        return

    if (getsize(infile) >> 20) < max_size:
        return

    with open(infile, 'r', encoding='utf8') as infile_h:
        text = infile_h.read()

        header_and_text = HEADER_END_PATTERN.split(text)
        header = header_and_text[0] + header_and_text[1]
        text = header_and_text[2].lstrip('\n')  # Ignore new lines after #META#Header#End#
        text = pre_clean_text(text)
        text = MISSING_DIRECTIONALITY_TAG_PATTERN.sub('\g<1>_ء_ \g<2>', text)

        paragraphs = text.split('\n\n')

        blocks = [(k, list(gr)) for k, gr in groupby(
            paragraphs,
            lambda s: bool(search(FIRST_LEVEL_HEADING_PATTERN, s))
        )]

        chunks = []

        aux = []
        for is_fst_header, paragraphs in blocks:
            if is_fst_header:
                #if debug and len(paragraphs) > 1:
                #    multiple_headings = "\n\n".join(paragraphs)
                #    print(f"\nfile {infile} has consecutive first level multiple_headings:"
                #          f"\n\n{multiple_headings}")
                if len(aux) > 1:
                    chunks.append("\n\n".join(aux))
                    aux = []
            for paragraph in paragraphs:
                aux.append(paragraph)

        if aux:
            chunks.append("\n\n".join(aux))

        # to avoid having too many small files
        chunks = resize_chunks(chunks)

        # there is only one first level heading in all the file
        if len(chunks) > 1:

            for i, chunk in enumerate(chunks, 1):

                final = header + '\n\n' + chunk
                if final[-1] != '\n':
                    final += '\n'

                outfile_path = get_part_filepath(file_path, i, file_ext)
                with open(outfile_path, 'w', encoding='utf8') as outfp:
                    outfp.write(final)

            check_file_split(infile, debug)


from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from pathlib import Path
from glob import glob
from functools import partial

from p_tqdm import p_uimap

from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600MIUAction
from eis1600.repositories.repo import TRAINING_DATA_REPO
from eis1600.toponyms.methods import toponym_category_annotation


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotate onomastic information in gold-standard MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument('-K', '--keep', action='store_true', help='Keep automatic tags (Ü-tags)')
    arg_parser.add_argument('-T', '--test', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='IDs or MIU file to process',
            action=CheckFileEndingEIS1600MIUAction
    )

    args = arg_parser.parse_args()
    debug = args.debug
    test = args.test
    keep = args.keep

    if args.input:
        toponym_category_annotation(args.input, test, keep)
    else:
        if test:
            with open(TRAINING_DATA_REPO + 'gold_standard.txt', 'r', encoding='utf-8') as fh:
                files_txt = fh.read().splitlines()

            infiles = [TRAINING_DATA_REPO + 'gold_standard_nasab/' + file for file in files_txt if Path(
                    TRAINING_DATA_REPO + 'gold_standard_nasab/' + file).exists()]
        else:
            infiles = glob(TRAINING_DATA_REPO + 'training_data_nasab_ML2/*.EIS1600')

        if debug:
            for file in infiles[:20]:
                print(file)
                toponym_category_annotation(file, test, keep)
        else:
            res = []
            res += p_uimap(partial(toponym_category_annotation, test=test, keep_automatic_tags=keep), infiles)

    print('Done')
